module Hobby
  module JSON
    module Keys
      class KeyParser
        def initialize &block
          @block = -> json, key {
            key = key.to_s if key.is_a? Symbol
            block.call json, key
          }
        end
        
        def [] key
          -> json {
            @block.call json, key
          }
        end
      end
    
      class ValueParser
        def initialize &block
          @block = block
        end
        
        def [] type
          -> value {
            @block.call value, type
          }
        end
      end
    
      require 'hobby/json'
      
      def self.included app
        app.include JSON
        app.extend Singleton
      
        %i[
          String
          Array
          Hash
        ].each do |symbol|
          app.type symbol do
            is_a Object.const_get symbol
            is_not_empty
          end
        end
      end
      
      def keys
        @keys ||= begin
                    self.class.keys
                      .transform_values { |parser| parser[json] }
                      .compact
                  rescue
                    response.status = 400
                    halt
                  end
      end
    
      module Singleton
        def keys
          @keys ||= {}
        end
        
        def key key, type = nil
          parser = if type
                     Both ValueMustExistFor[key], TypeOfValueMustBe[type]
                   else
                     ValueMustExistFor[key]
                   end
        
          self.keys[key] = parser
        end
        
        def optional_key key, type = nil
          parser = if type
                     Both ValueMayExistFor[key], TypeOfValueMustBe[type]
                   else
                     ValueMayExistFor[key]
                   end
        
          self.keys[key] = parser
        end
        
        
        def Both key_parser, value_parser
          -> json {
            if value = key_parser[json]
              value_parser[value]
            end
          }
        end
        
        
        ValueMustExistFor = KeyParser.new do |json, key|
          value = json[key]
          fail unless value
          value
        end
        
        ValueMayExistFor = KeyParser.new do |json, key|
          json[key]
        end
        
        TypeOfValueMustBe = ValueParser.new do |value, type|
          fail unless type === value
          if [String, Array, Hash].include? type
            fail if value.empty?
          end
          value
        end
        
        
        def optional &block
          new_class = dup
          new_class.define_singleton_method :key, &method(:optional_key)
          new_class.instance_exec &block
          self.keys.merge! new_class.keys
        end
        
        def initialize_copy
          super
          @keys = {}
        end
        
        
        def types
          @types ||= {}
        end
        
        def type symbol, &default
          types[symbol] = Type.new &default
          define_singleton_method symbol do |&custom|
            types[symbol].expand &custom
          end
        end
      end
    
      class Type
        def initialize &block
          @gates = {}
          instance_exec &block
        end
        
        attr_reader :gates
        
        def expand &block
          new_type = dup
          new_type.instance_exec &block
          new_type
        end
        
        def initialize_copy original
          super
          @gates = original.gates.dup
        end
        
        def is_a type
          @gates[__callee__] = -> it {
            it.is_a? type
          }
        end
        
        def is_not_empty
          @gates[__callee__] = -> it {
            not it.empty?
          }
        end
        
        def may_be_empty
          @gates.delete :is_not_empty
        end
        
        def size value
          @gates[__callee__] = -> it {
            value === it.size
          }
        end
        
        def === it
          @gates.values.each do |gate|
            fail unless gate === it
          end
        end
      end
    end
  end
end
